@@ -1,5 +1,5 @@
 ## FileSystemModel.py - FileSystemModel class
-## Copyright (C) 2005 Tony Tsui <tsui.tony@gmail.com>
+## Copyright (C) 2005-2006 Tony Tsui <tsui.tony@gmail.com>
 ##
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -31,16 +31,128 @@ from util import *
 # Set to True for debug output
 DEBUG = False
 
+def debugLocking(action):
+	if DEBUG is False:
+		return
+	
+	caller = sys._getframe(1)
+	try:
+		func_name = inspect.getframeinfo(caller)[2]
+	finally:
+		del caller
+
+	print func_name + '(): ' + action
+
 class FileSystemModel(gtk.ListStore):
 	TYPE_COL, ICON_COL, NAME_COL, DATE_COL, SIZE_COL = range(5)
 	LIST_TYPES = []
 	LIST_TYPES.insert(TYPE_COL, gobject.TYPE_STRING)
-	LIST_TYPES.insert(ICON_COL, gobject.TYPE_STRING)
+	LIST_TYPES.insert(ICON_COL, gtk.gdk.Pixbuf)
 	LIST_TYPES.insert(NAME_COL, gobject.TYPE_STRING)
 	LIST_TYPES.insert(DATE_COL, gobject.TYPE_STRING)
 	LIST_TYPES.insert(SIZE_COL, gobject.TYPE_STRING)
 
-	def __init__(self, show_parent_dir=False):
+	# Flag to indicate icons have already been loaded
+	icons_loaded = False
+
+	# Set in FileSystemModel::__load_icons()
+	file_icon = None
+	dir_icon = None
+	video, audio, tap, tap_ini = ('.rec', '.mp3', '.tap', '.ini')
+	icon_dict = { video : None,
+	              audio : None,
+	              tap : None,
+	              tap_ini : None,
+	            }
+	
+	icon_theme = gtk.icon_theme_get_default()
+	theme_change_callback = []
+	
+	@staticmethod
+	def __connect_theme_change_callback(callback):
+		""" Register theme change callback functions from each FileSystemModel
+		    instance.
+		    
+		    Typically the callback functions will refresh the user interface
+		    with the new icons.
+		"""
+		FileSystemModel.theme_change_callback.append(callback)
+		
+	@staticmethod
+	def __load_icons(icon_theme, datadir):
+		""" Load icons for files, directories, video and audio files.
+		
+		    This function should be called once on class initialisation and
+		    whenever a theme change occurs.
+		"""
+		if not FileSystemModel.icons_loaded:
+			icon_theme.connect('changed', FileSystemModel.__load_icons, datadir)
+		
+		img = gtk.Image()
+		FileSystemModel.file_icon = img.render_icon(gtk.STOCK_FILE,
+		                                            gtk.ICON_SIZE_LARGE_TOOLBAR)
+	
+		settings = gtk.settings_get_for_screen(img.get_screen())
+		icon_size = gtk.icon_size_lookup_for_settings(settings,
+		                                            gtk.ICON_SIZE_LARGE_TOOLBAR)
+		if icon_size:
+			icon_size = max(icon_size[0], icon_size[1])
+		else:
+			icon_size = 24
+
+		# List of icons and their icon theme name
+		icons = { 'dir_icon' : 'folder',
+		          'video_icon' : 'video-x-generic',
+		          'audio_icon' : 'audio-x-generic',
+		          'tap_icon' : 'application-x-executable',
+		        }
+	
+		# Try and load icon from icon theme	
+		for key, icon_name in icons.iteritems():
+			try:
+				icons[key] = icon_theme.load_icon(icon_name, icon_size,
+				                                  gtk.ICON_LOOKUP_USE_BUILTIN)
+			except gobject.GError, exc:
+				icons[key] = None
+				pass
+
+		FileSystemModel.dir_icon = icons['dir_icon']
+		FileSystemModel.icon_dict[FileSystemModel.video] = icons['video_icon']
+		FileSystemModel.icon_dict[FileSystemModel.audio] = icons['audio_icon']
+		FileSystemModel.icon_dict[FileSystemModel.tap] = icons['tap_icon']
+		FileSystemModel.icon_dict[FileSystemModel.tap_ini] = icons['tap_icon']
+
+		# Load default icons if icon not found in theme
+		if FileSystemModel.dir_icon == None:
+			FileSystemModel.dir_icon = img.render_icon(gtk.STOCK_DIRECTORY,
+		                                               gtk.ICON_SIZE_LARGE_TOOLBAR)
+		
+		if FileSystemModel.icon_dict[FileSystemModel.video] == None:	
+			FileSystemModel.icon_dict[FileSystemModel.video] = gtk.gdk.pixbuf_new_from_file(datadir + '/' + 'video.png')
+
+		if FileSystemModel.icon_dict[FileSystemModel.audio] == None:	
+			FileSystemModel.icon_dict[FileSystemModel.audio] = gtk.gdk.pixbuf_new_from_file(datadir + '/' + 'audio.png')
+
+		if FileSystemModel.icon_dict[FileSystemModel.tap] == None:	
+			FileSystemModel.icon_dict[FileSystemModel.tap] = gtk.gdk.pixbuf_new_from_file(datadir + '/' + 'tap.png')
+			FileSystemModel.icon_dict[FileSystemModel.tap_ini] = FileSystemModel.icon_dict[FileSystemModel.tap]
+		
+		# Call registered theme change callback functions
+		if FileSystemModel.icons_loaded:
+			for callback in FileSystemModel.theme_change_callback:
+				callback()
+		
+		FileSystemModel.icons_loaded = True
+	
+	@staticmethod
+	def getIconForFile(file):
+		extension = file[-4:].lower()
+		if FileSystemModel.icon_dict.has_key(extension):
+			return FileSystemModel.icon_dict[extension]
+		else:
+			return FileSystemModel.file_icon
+	
+	def __init__(self, datadir, show_parent_dir=False):
 		self.current_dir = None
 		self.show_parent_dir = show_parent_dir
 		gtk.ListStore.__init__(self, FileSystemModel.LIST_TYPES[FileSystemModel.TYPE_COL],
@@ -49,9 +161,16 @@ class FileSystemModel(gtk.ListStore):
 							   FileSystemModel.LIST_TYPES[FileSystemModel.DATE_COL],
 			                   FileSystemModel.LIST_TYPES[FileSystemModel.SIZE_COL])
 
+		# Load icons once
+		if not FileSystemModel.icons_loaded:
+			FileSystemModel.__load_icons(FileSystemModel.icon_theme, datadir)
+			
+		# Connect callback function for theme change
+		FileSystemModel.__connect_theme_change_callback(self.changeDir)
+
 	def abspath(self, file):
-		if self.current_dir[-1] != self.separator and file[0] != self.separator:
-			return self.current_dir + self.separator + file
+		if self.current_dir[-1] != self.slash and file[0] != self.slash:
+			return self.current_dir + self.slash + file
 		else:
 			return self.current_dir + file
 		
@@ -119,11 +238,11 @@ class FileSystemModel(gtk.ListStore):
 	def date_sort_func(self, model, iter1, iter2, col=None):
 		# Date column is empty for '..' directory
 		date1_str = model.get_value(iter1, col)
-		if len(date1_str) == 0:
+		if date1_str == None or len(date1_str) == 0:
 			return -1
 		
 		date2_str = model.get_value(iter2, col)
-		if len(date2_str) == 0:
+		if date2_str == None or len(date2_str) == 0:
 			return 1
 
 		format = '%a %b %d %Y'
@@ -138,29 +257,40 @@ class FileSystemModel(gtk.ListStore):
 			return 1
 
 	def size_sort_func(self, model, iter1, iter2, col=None):
-		size1 = model.get_value(iter1, col).split()
-		size2 = model.get_value(iter2, col).split()
+		size1 = model.get_value(iter1, col)
+		size2 = model.get_value(iter2, col)
 
+		# We can get None on PVR FS.
+		if size1 == None and size2 != None:
+			return -1
+		elif size1 != None and size2 == None:
+			return 1
+		elif size1 == None and size2 == None:
+			return self.string_sort_func(model, iter1, iter2, FileSystemModel.NAME_COL);
+
+		# Handle when size is empty string. Typical for directories on PC.			
 		size1_len = len(size1)
 		size2_len = len(size2)
-		
-		if size1_len < size2_len:
+
+		if size1_len == 0 and size2_len > 0:
 			return -1
-		elif size1_len > size2_len:
+		elif size1_len > 0 and size2_len == 0:
 			return 1
-		else:
-			return 0
+		elif size1_len == 0 and size2_len == 0:
+			return self.string_sort_func(model, iter1, iter2, FileSystemModel.NAME_COL);
+		
+		size1 = size1.split()
+		size2 = size2.split()
 
 		unit_weight = { 'B' : 1, 'KB' : 2, 'MB' : 3, 'GB' : 4 }
 
-		# Sizes in bytes may not have 'B' unit postfix
+		# Sizes in bytes may not have 'B' unit
 		if len(size1) < 2:
 			size1[1] = unit_weight['B']
 		else:
 			size1[1] = unit_weight[size1[1]]
 
-		return 1
-		# Sizes in bytes may not have 'B' unit appended
+		# Sizes in bytes may not have 'B' unit
 		if len(size2) < 2:
 			size2[1] = unit_weight['B']
 		else:
@@ -181,18 +311,20 @@ class FileSystemModel(gtk.ListStore):
 			return 1
 
 class PCFileSystemModel(FileSystemModel):
-	def __init__(self, show_parent_dir=False):
-		FileSystemModel.__init__(self, show_parent_dir)
+	def __init__(self, datadir, cwd=None, show_parent_dir=False):
+		FileSystemModel.__init__(self, datadir, show_parent_dir)
+
+		self.slash = '/'
 		
-		# FIXME: Get dir from when Guppy last exited
-		self.current_dir = os.environ['HOME']
-		
-		self.separator = '/'
+		if cwd and self.exists(cwd):
+			self.current_dir = cwd
+		else:
+			self.current_dir = os.environ['HOME']		
 		
 	def changeDir(self, dir=None):
 		if dir:
 			# Append CWD if dir is not an absolute path
-			if dir[0] != self.separator:
+			if dir[0] != self.slash:
 				dir = self.abspath(dir)
 		else:
 			dir = self.current_dir
@@ -200,7 +332,7 @@ class PCFileSystemModel(FileSystemModel):
 		dir = os.path.normpath(dir)
 		# os.path.normpath() doesn't normalise '//' to '/'
 		if dir == '//':
-			dir = '/'
+			dir = self.slash
 		
 		if not os.access(dir, os.F_OK):
 			return False
@@ -213,21 +345,20 @@ class PCFileSystemModel(FileSystemModel):
 		# Parent directory
 		if self.show_parent_dir:
 			self.append(['d', gtk.STOCK_DIRECTORY, '..', '', ''])
-		
+
 		for file in os.listdir(self.current_dir):
 			try:
 				mode = os.stat(self.abspath(file))
 			except OSError, (errno, strerror):
-				print 'PCFileSystemModel::changeDir(%s): OSError(%s): %s\n' % (self.abspath(file), errno, strerror)
 				continue
 
 			if stat.S_ISDIR(mode[stat.ST_MODE]):
 				type = 'd'
-				icon = gtk.STOCK_DIRECTORY
+				icon = FileSystemModel.dir_icon
 				size = ''
 			else:
 				type = 'f'
-				icon = gtk.STOCK_FILE
+				icon = FileSystemModel.getIconForFile(file)
 				size = humanReadableSize(mode[stat.ST_SIZE])
 			
 			mtime = time.strftime('%a %b %d %Y', time.localtime(mode[stat.ST_MTIME]))
@@ -247,7 +378,10 @@ class PCFileSystemModel(FileSystemModel):
 			os.remove(path)
 		
 	def exists(self, file):
-		return os.access(self.abspath(file), os.F_OK)
+		if file[0] != self.slash:
+			file = self.abspath(file)
+		
+		return os.access(file, os.F_OK)
 		
 	def freeSpace(self):
 		cmd = 'df -P ' + '"' + self.current_dir + '"'
@@ -288,17 +422,19 @@ class PCFileSystemModel(FileSystemModel):
 		os.rename(self.abspath(old), self.abspath(new))
 
 class PVRFileSystemModel(FileSystemModel):
-	def __init__(self, show_parent_dir=False):
-		FileSystemModel.__init__(self, show_parent_dir)
+	def __init__(self, datadir, cwd=None, show_parent_dir=False):
+		FileSystemModel.__init__(self, datadir, show_parent_dir)
 
-		self.separator = '\\'
-		# FIXME: Get dir from when Guppy last exited
-		# We need to use an empty string to list the PVR root directory.
-		self.current_dir = self.separator
+		self.slash = '\\'
 
-		self.freespace = 0
-				
 		self.puppy = Puppy()
+		
+		if cwd and self.exists(cwd):
+			self.current_dir = cwd
+		else:		
+			self.current_dir = self.slash
+
+		self.freespace = 0				
 
 		self.dir_tree_lock = threading.Lock()
 
@@ -310,40 +446,42 @@ class PVRFileSystemModel(FileSystemModel):
 		"""
 		if dir:
 			# Append CWD if dir is not an absolute path
-			if dir[0] != self.separator:
+			if dir[0] != self.slash:
 				dir = self.abspath(dir)
 		else:
 			dir = self.current_dir
 
-		norm_path = os.path.normpath(dir.replace(self.separator, '/'))
+		norm_path = os.path.normpath(dir.replace(self.slash, '/'))
 		# norm_path is '.' when _dir_ is an empty string. _dir_ is an empty
 		# string for the PVR root directory.
 		# Also don't change dir when changing to '..' directory.
 		if norm_path == '.' or norm_path == '..':
-			dir = self.separator
+			dir = self.slash
 		else:
-			dir = norm_path.replace('/', self.separator)
+			dir = norm_path.replace('/', self.slash)
 
 		dir_node = self.findDirectory(dir)
 		if dir_node == None:
 			return
 		
 		self.current_dir = dir
-
+		
 		# Clear model
 		if len(self) > 0:
 			self.clear()
 
-		for file_name, file_info in dir_node.getFiles():  
-			file_info.insert(FileSystemModel.ICON_COL, gtk.STOCK_FILE)
+		for file_name, file_info in dir_node.getFiles():
+			icon = FileSystemModel.getIconForFile(file_name)
+								
+			file_info.insert(FileSystemModel.ICON_COL, icon)
 		
 			file_info[FileSystemModel.SIZE_COL] = humanReadableSize(int(file_info[FileSystemModel.SIZE_COL]))
 			
 			self.append(file_info)
 			
 		for dir, dir_info in dir_node.getDirectories():
-			dir_info.insert(FileSystemModel.ICON_COL, gtk.STOCK_DIRECTORY)
-		
+			dir_info.insert(FileSystemModel.ICON_COL, FileSystemModel.dir_icon)
+			
 			dir_info[FileSystemModel.SIZE_COL] = humanReadableSize(int(dir_info[FileSystemModel.SIZE_COL]))
 			
 			self.append(dir_info)
@@ -353,11 +491,21 @@ class PVRFileSystemModel(FileSystemModel):
 			self.append([ 'd', gtk.STOCK_DIRECTORY, '..', '', ''])
 
 	def delete(self, file):
-		print 'file = ', file
 		self.puppy.delete(self.abspath(file))
 
 	def exists(self, file):
-		pvr_files = self.puppy.listDir(self.current_dir)
+		# puppy.listDir() can not list files so instead we list the parent
+		# directory and search for the file in there.
+		
+		# Get parent directory
+		if file[0] == self.slash:
+			last_slash = file.rfind(self.slash) 
+			parent_dir = file[:last_slash]
+			file = file[last_slash+1:]
+		else:
+			parent_dir = self.current_dir
+
+		pvr_files = self.puppy.listDir(parent_dir)
 		
 		for p_file in pvr_files:
 			if p_file[1] == file:
@@ -370,15 +518,20 @@ class PVRFileSystemModel(FileSystemModel):
 		
 		Returns: DirectoryNode object
 		"""
+		debugLocking('locking')
 		self.dir_tree_lock.acquire()
 		cur_node = self.dir_tree
 		
-		for dir in path.split('\\'):
+		# Strip trailing slash
+		path = path.rstrip(self.slash)
+		
+		for dir in path.split(self.slash):
 			if len(dir) == 0:
 				continue
 			cur_node, node_info = cur_node.getDirectory(dir)
 			if cur_node == None:
 				break
+		debugLocking('unlocking')
 		self.dir_tree_lock.release()
 
 		return cur_node
@@ -441,6 +594,7 @@ class PVRFileSystemModel(FileSystemModel):
 		if DEBUG:
 			print 'Start updateCache()'
 			
+		debugLocking('locking')
 		self.dir_tree_lock.acquire()
 
 		new_dir_tree = None
@@ -450,12 +604,14 @@ class PVRFileSystemModel(FileSystemModel):
 			try:
 				new_dir_tree = self.scanDirectory('')
 			except PuppyBusyError:
-				print 'updateCache(): Exception: PuppyBusyError'
 				# Sleep for 1 second before trying again
 				time.sleep(1)
+				continue
 			except PuppyError:
+				debugLocking('unlocking')
 				self.dir_tree_lock.release()
 				raise
+			break
 
 		# Failed to update cache. Use existing cache.
 		if new_dir_tree == None:
@@ -463,11 +619,77 @@ class PVRFileSystemModel(FileSystemModel):
 			
 		self.dir_tree = new_dir_tree
 
+		debugLocking('unlocking')
 		self.dir_tree_lock.release()
 		
 		if DEBUG:
 			print 'End updateCache()'
 
+	def updateDirectory(self, dir):
+		"""Update the directory node in the directory tree.
+		
+		"""
+		if DEBUG:
+			print 'Start updateDirectory(%s)' % (dir)
+			
+		# Strip trailing slash
+		dir = dir.rstrip(self.slash)
+
+		debugLocking('locking')
+		self.dir_tree_lock.acquire()
+
+		# Call scanDirectory() directly if dir is root dir, i.e. '\'
+		if len(dir) == 0:
+			self.dir_tree = self.scanDirectory('')
+			debugLocking('unlocking')
+			self.dir_tree_lock.release()
+			if DEBUG:
+				print 'End updateDirectory(\\)'
+			return
+
+		# Get components of directory path as a list
+		path = dir.split('\\')
+
+		# Find parent node		
+		parent_node = self.dir_tree
+		for name in path[:-1]:
+			if len(name) == 0:
+				continue
+			par_node, node_info = parent_node.getDirectory(name)
+			if parent_node == None:
+				break
+				
+		cur_node, node_info = parent_node.getDirectory(path[-1])
+		
+		new_node = None
+		# Attempt to update cache twice in case puppy was busy the first time
+		for i in xrange(2):
+			try:
+				new_node = self.scanDirectory(dir)
+			except PuppyBusyError:
+				# Sleep for 1 second before trying again
+				time.sleep(1)
+				continue
+			except PuppyError:
+				debugLocking('unlocking')
+				self.dir_tree_lock.release()
+				raise
+			break
+
+		# Failed to update cache. Use existing node.
+		if new_node == None:
+			new_node = cur_node
+
+		# Replace wit new node
+		parent_node.addDirectory(new_node, node_info)
+		
+		debugLocking('unlocking')
+		self.dir_tree_lock.release()
+		
+		if DEBUG:
+			print 'End updateDirectory()'
+		
+
 class DirectoryNode:
 	def __init__(self, name):
 		self.name = name
@@ -511,3 +733,10 @@ class DirectoryNode:
 	def getFiles(self):
 		return copy.deepcopy(self.files)
 		
+	def print_tree(self):
+		print '\t' + self.name
+		for node, info in self.sub_directories.itervalues():
+			node.print_tree()
+			
+		for file in self.files:
+			print file
