@@ -20,7 +20,7 @@ import os
 import stat
 import threading
 import Queue
-import datetime
+import time
 
 import gtk
 import gtk.glade
@@ -72,8 +72,8 @@ class GuppyWindow:
 		
 		# Connect buttons in toolbar to their respective action
 		for action, widget in [ (actions['turbo'], 'turbo_btn'),
-		                (actions['upload'], 'upload_btn'),
-						(actions['download'], 'download_btn') ]:
+		                        (actions['upload'], 'upload_btn'),
+		                        (actions['download'], 'download_btn') ]:
 			btn = self.glade_xml.get_widget(widget)
 			action.connect_proxy(btn)
 		
@@ -98,8 +98,12 @@ class GuppyWindow:
 		else:
 			show_parent_dir = False
 
-		self.pvr_model = PVRFileSystemModel(show_parent_dir)
-		self.pc_model = PCFileSystemModel(show_parent_dir)
+		self.pc_model = PCFileSystemModel(self.datadir,
+		                                  cwd=getConfValue('PC_CWD'),
+		                                  show_parent_dir=show_parent_dir)
+		self.pvr_model = PVRFileSystemModel(self.datadir,
+		                                    cwd=getConfValue('PVR_CWD'),
+		                                    show_parent_dir=show_parent_dir)
 		
 		self.pc_path_entry_box = self.glade_xml.get_widget('pc_path_entry_box')
 		self.pvr_path_entry_box = self.glade_xml.get_widget('pvr_path_entry_box')
@@ -144,10 +148,11 @@ class GuppyWindow:
 		self.transfer_thread.start()
 	
 		self.turbo = False
+		self.turbo_warn = True
 		
 		self.last_file_transfer = None
 		self.quit_after_transfer = False
-		
+
 	def initUIManager(self):
 		"""Initialise the UIManager.
 		
@@ -263,6 +268,7 @@ class GuppyWindow:
 			sort_func = fs_model.sort_func
 			liststore.set_default_sort_func(sort_func, FileSystemModel.NAME_COL)
 
+			treeview.set_search_column(FileSystemModel.NAME_COL)
 			treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
 			
 			treeview.connect('row-activated',
@@ -280,6 +286,10 @@ class GuppyWindow:
 			                 self.on_treeview_button_press,
 			                 fs_model)
 
+			treeview.connect('key-press-event',
+			                 self.on_treeview_key_press,
+			                 fs_model)
+
 			# Text cell for file name			
 			text_cell = gtk.CellRendererText()
 			text_cell.set_property('ellipsize', pango.ELLIPSIZE_END)
@@ -296,7 +306,7 @@ class GuppyWindow:
 			col.pack_start(pixb_cell, False)
 			col.pack_start(text_cell, True)
 			col.set_attributes(text_cell, text=FileSystemModel.NAME_COL)
-			col.set_attributes(pixb_cell, stock_id=FileSystemModel.ICON_COL)
+			col.set_attributes(pixb_cell, pixbuf=FileSystemModel.ICON_COL)
 			col.set_clickable(True)
 			col.set_resizable(True)
 			col.set_sort_indicator(True)
@@ -334,6 +344,26 @@ class GuppyWindow:
 	
 		Return: True if all files deleted.
 		"""
+		msg = '<b>' + _('Are you sure you want to permanently delete') + '\n'
+		
+		files_len =  len(files)
+		if files_len > 1:
+			msg += ' ' + _('the') + ' ' + str(files_len) + ' ' + _('selected items?')
+		else:
+			msg += ' "' + files[0] + '"?'
+			
+		msg += '</b>\n\n' + _('If you delete an item, it is permanently lost.')
+		
+		dialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION, buttons=gtk.BUTTONS_CANCEL)
+		dialog.set_markup(msg)
+		dialog.add_button(gtk.STOCK_DELETE, 0)
+		
+		response = dialog.run()
+		dialog.destroy()
+
+		if response == gtk.RESPONSE_CANCEL or response == gtk.RESPONSE_DELETE_EVENT:
+			return False
+		
 		retval = True
 		for name in files:
 			deleted = False
@@ -359,14 +389,15 @@ class GuppyWindow:
 				
 					# Try to delete again
 					continue
+				except PuppyBusyError:
+					self.showNoFileOperationDialog()
+					return
 				except PuppyError, error:
 					self.pvr_error_btn.show()
 					self.pvr_error_window.addError(_('Failed to delete') + ' ' + name, error)
-					
-					deleted = True
-					continue
+					break
 
-		self.updateTreeViews(fs_model)
+		self.updateTreeViews(fs_model, True)
 		self.updateFreeSpace(fs_model)
 		
 		return retval
@@ -470,15 +501,18 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 			response = dialog.run()
 			dialog.destroy()
 
+			return
+		except PuppyBusyError:
+			self.showNoFileOperationDialog()
 			return
 		except PuppyError, error:
 			self.pvr_error_btn.show()
 			self.pvr_error_window.addError(_('Failed to make a folder'), error)
-		
+			return
 		
 		# Update model to get new folder
-		self.updateTreeViews(fs_model)
-
+		self.updateTreeViews(fs_model, True)
+		
 		# Get row for new folder		
 		model = treeview.get_model()
 		for row in model:
@@ -549,8 +583,9 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 			except PuppyError, error:
 				self.pvr_error_btn.show()
 				self.pvr_error_window.addError(_('Failed to rename') + ' ' + old_name, error)
+				break
 		
-		self.updateTreeViews(fs_model)
+		self.updateTreeViews(fs_model, True)
 
 	def on_name_cell_editing_cancelled(self, cell, data=None):
 		# Set editable to False so users can't edit the cell by clicking on it.
@@ -618,6 +653,11 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 		self.updateTreeViews()
 		
 	def on_rename_btn_clicked(self, widget, treeview, fs_model):
+		# Fail rename on PVR if a transfer is in progress
+		if isinstance(fs_model, PVRFileSystemModel) and self.puppy.isActive():
+			self.showNoFileOperationDialog()
+			return
+		
 		selection = treeview.get_selection()
 		model, files = selection.get_selected_rows()
 		
@@ -639,7 +679,7 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 		self.show_hidden = not self.show_hidden
 		self.pc_liststore.get_model().refilter()
 		self.pvr_liststore.get_model().refilter()
-
+	
 	def on_treeview_button_press(self, treeview, event, fs_model):
 		if event.button == 3:
 			time = event.time
@@ -650,27 +690,25 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 				self.file_popup_rename_btn.set_sensitive(True)
 			else:
 				self.file_popup_rename_btn.set_sensitive(False)
+				
+			if len(files) >= 1:
+				self.file_popup_delete_btn.set_sensitive(True)
+			else:
+				self.file_popup_delete_btn.set_sensitive(False)
+
+			btns = [ (self.file_popup_delete_btn, self.on_delete_btn_clicked),
+			         (self.file_popup_rename_btn, self.on_rename_btn_clicked),
+			         (self.file_popup_mkdir_btn, self.on_mkdir_btn_clicked) ]
 
 			# Connect signal handler for each popup menuitem with the treeview
 			# and fs_model where the popup menu appeared.
-			handler_id = self.file_popup_delete_btn.connect('activate',
-			                                                self.on_delete_btn_clicked,
-			                                                treeview,
-			                                                fs_model)
-			self.file_popup_delete_btn.set_data('handler_id', handler_id)
-			
-			handler_id = self.file_popup_mkdir_btn.connect('activate',
-			                                               self.on_mkdir_btn_clicked,
-			                                               treeview,
-			                                               fs_model)
-			self.file_popup_mkdir_btn.set_data('handler_id', handler_id)
-
-			handler_id = self.file_popup_rename_btn.connect('activate',
-			                                                self.on_rename_btn_clicked,
-			                                                treeview,
-			                                                fs_model)
-			self.file_popup_rename_btn.set_data('handler_id', handler_id)
-
+			for btn, callback in btns:
+				handler_id = btn.connect('activate',
+				                         callback,
+				                         treeview,
+				                         fs_model)
+				btn.set_data('handler_id', handler_id)
+				
 			self.file_popup.popup( None, None, None, event.button, time)
 			
 			# Return True so row selection doesn't get changed.
@@ -712,6 +750,18 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 			actiongrp.set_sensitive(True)
 		else:
 			actiongrp.set_sensitive(False)
+
+	def on_treeview_key_press(self, treeview, event, fs_model):
+		
+		# <F2>: Rename selected file.
+		if event.keyval == 65471:
+			selection = treeview.get_selection()
+			model, files = selection.get_selected_rows()
+			if len(files) == 1:
+				self.on_rename_btn_clicked(None, treeview, fs_model)
+		# <Delete>: Delete selected files.
+		elif event.keyval == 65535:
+			self.on_delete_btn_clicked(None, treeview, fs_model)
 	
 	def on_treeview_row_activated(self, widget, path, col, fs_model):
 		model = widget.get_model()
@@ -766,7 +816,24 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 
 	def on_turbo_toggled(self, widget, data=None):
 		self.turbo = widget.get_active()
+		
+		if self.turbo and self.turbo_warn:
+			dialog = gtk.MessageDialog(message_format=_('You can not use your PVR remote control when Turbo mode is on.'),
+			                           type=gtk.MESSAGE_WARNING,
+			                           buttons=gtk.BUTTONS_CLOSE)
 
+			checkbox = gtk.CheckButton(_("Warn me when I enable Turbo mode"))
+			checkbox.set_active(self.turbo_warn)
+			checkbox.connect('toggled', self.on_turbo_warn_toggled)
+			checkbox.show()
+
+			dialog.vbox.pack_end(checkbox)
+			dialog.run()
+			dialog.destroy()
+
+	def on_turbo_warn_toggled(self, widget):
+		self.turbo_warn = widget.get_active()
+		
 	def on_upload_btn_clicked(self, widget, data=None):
 		self.transferFile('upload')
 
@@ -777,7 +844,18 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 		return True
 
 	def reallyQuit(self):
+		# Save current working directory path
+		setConfValue('PC_CWD', self.pc_model.getCWD())
+		setConfValue('PVR_CWD', self.pvr_model.getCWD())
+		
 		gtk.main_quit()
+
+	def showNoFileOperationDialog(self):
+		dialog = gtk.MessageDialog(message_format=_('It is not possible to delete, rename, or create folders on the PVR during a file transfer.'),
+		                           type=gtk.MESSAGE_ERROR,
+		                           buttons=gtk.BUTTONS_CLOSE)
+		dialog.run()
+		dialog.destroy()
 		
 	def show_transfer_frame(self):
 		# Set position of pane separator
@@ -839,6 +917,7 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 		for path in files:
 			iter = model.get_iter(path)
 			file = model.get_value(iter, FileSystemModel.NAME_COL)
+			date = model.get_value(iter, FileSystemModel.DATE_COL)
 
 			if direction == 'download':
 				src_dir = self.pvr_model.getCWD()
@@ -869,7 +948,8 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 			from_label.set_text(src_dir)
 			to_label.set_text(dst_dir)
 	
-			file_transfer = FileTransfer(direction, src_file, dst_file, xml)
+			file_transfer = FileTransfer(direction, src_file, dst_file,
+			                             date, xml)
 
 			# Connect transfer instance remove button signal handler
 			remove_btn = xml.get_widget('transfer_remove_button')
@@ -967,7 +1047,7 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 		return True
 	
 
-	def updateTreeViews(self, fs_model=None):
+	def updateTreeViews(self, fs_model=None, cur_dir_only=False):
 		# Update FileSystemModel view
 		if fs_model == self.pc_model:
 			models = [ (self.pc_model, self.pc_treeview) ]
@@ -988,7 +1068,10 @@ You can download Puppy from <i>http://sourceforge.net/projects/puppy</i>'''))
 			# Update PVR file system cache
 			if isinstance(model, PVRFileSystemModel):
 				try:
-					model.updateCache()
+					if cur_dir_only:
+						model.updateDirectory(model.getCWD())
+					else:
+						model.updateCache()
 				except PuppyNoPVRError:
 					raise
 				except PuppyError, error:
@@ -1020,6 +1103,7 @@ class PVRErrorWindow:
 		self.glade_file = glade_file		
 		self.error_window = glade_xml.get_widget('pvr_error_window')
 		self.pvr_error_output = glade_xml.get_widget('pvr_error_output')
+		self.pvr_error_box = glade_xml.get_widget('pvr_error_box')
 
 		self.liststore = gtk.ListStore(PVRErrorWindow.LIST_TYPES[PVRErrorWindow.ICON_COL],
 		                               PVRErrorWindow.LIST_TYPES[PVRErrorWindow.ICON_SIZE_COL],
@@ -1042,11 +1126,20 @@ class PVRErrorWindow:
 		self.pvr_error_treeview.append_column(col)
 		
 	def addError(self, msg, error_output):
-		error_msg = _('ERROR') + ': ' + datetime.datetime.now().strftime('%a %b %d, %I:%M %p') +'\n'
+		error_msg = _('ERROR') + ': ' + time.strftime('%a %b %d, %I:%M %p', time.localtime()) +'\n'
 		error_msg += '<b>' + msg + '</b>'
 		
 		self.liststore.append([gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_DIALOG, error_msg, error_output])
 
+	def on_expander_activate(self, widget):
+		if widget.get_property('expanded'):
+			box_expand = False
+		else:
+			box_expand = True
+			
+		self.pvr_error_box.set_child_packing(widget, box_expand, True, 0,
+		                                     gtk.PACK_START)
+		
 	def on_pvr_error_treeview_changed(self, widget):
 		model, iter = widget.get_selected()
 		if iter == None:
@@ -1070,11 +1163,12 @@ class PVRErrorWindow:
 		self.error_window.show()
 
 class FileTransfer:
-	def __init__(self, direction, src, dst, xml):
+	def __init__(self, direction, src, dst, file_date, xml):
 		self.direction = direction
 		self.src = src
 		self.dst = dst
 		self.xml = xml
+		self.file_date = file_date
 		
 		self.alive = True
 		self.quit_after_transfer = False
@@ -1103,7 +1197,7 @@ class TransferThread(threading.Thread):
 		self.guppy = guppy
 		self.file_queue = self.guppy.transfer_queue
 		self.complete_queue = self.guppy.transfer_complete_queue
-
+		
 	def run(self):
 		while True:
 				
@@ -1166,7 +1260,7 @@ class TransferThread(threading.Thread):
 			percent = True			
 			while percent != None:
 				try:
-					percent, speed, time = self.guppy.puppy.getProgress()
+					percent, speed, transfer_time = self.guppy.puppy.getProgress()
 				except PuppyError, error:
 					# Quit trying to transfer file after a certain number of attempts
 					if transfer_attempt > TransferThread.NUM_OF_TRANSFER_ATTEMPTS:
@@ -1199,7 +1293,7 @@ class TransferThread(threading.Thread):
 					
 				gtk.gdk.threads_enter()
 				progress_bar.set_fraction(float(percent)/100)
-				progress_bar.set_text('(' + time['remaining'] + ' ' + _('Remaining') + ')')
+				progress_bar.set_text('(' + transfer_time['remaining'] + ' ' + _('Remaining') + ')')
 
 				while gtk.events_pending():
 					gtk.main_iteration()
@@ -1209,6 +1303,18 @@ class TransferThread(threading.Thread):
 			# turbo mode is left on.
 			if self.guppy.turbo == True:
 				self.guppy.puppy.setTurbo(False)
+
+			# Set modification time of downloaded file to the same time as the
+			# as on the PVR.
+			if transfer_successful and direction == 'download':
+				# Parse date string
+				time_struct = time.strptime(file_transfer.file_date, '%a %b %d %Y')
+				
+				# Convert to seconds since the epoch
+				time_secs = time.mktime(time_struct)
+				
+				# Set modification time
+				os.utime(dst_file, (int(time.time()), time_secs))
 			
 			gtk.gdk.threads_enter()
 			if transfer_successful:
@@ -1218,7 +1324,13 @@ class TransferThread(threading.Thread):
 			else:
 				progress_bar.set_text(_('Transfer Failed'))
 				self.guppy.pvr_error_btn.show()
-				self.guppy.pvr_error_window.addError(_('Transfer Failed'), transfer_error)
+
+				if direction == 'download':
+					msg = _('Failed to download:') + '\n' + src_file
+				else:
+					msg = _('Failed to upload: ') + '\n' + src_file
+
+				self.guppy.pvr_error_window.addError(msg, transfer_error)
 
 			gtk.gdk.threads_leave()
 				
@@ -1233,9 +1345,6 @@ class TransferThread(threading.Thread):
 			# Swap Stop button for Remove button
 			gtk.gdk.threads_enter()
 			stop_btn.hide()
-			gtk.gdk.threads_leave()
-			
-			gtk.gdk.threads_enter()
 			remove_btn.show()
 			gtk.gdk.threads_leave()
 
