@@ -16,12 +16,12 @@
 ## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 APP_NAME = 'Guppy'
-VERSION = '0.0.90'
+VERSION = '0.0.91'
 DESCRIPTION='Topfield PVR File Downloader'
 AUTHOR = 'Tony Tsui'
 AUTHOR_EMAIL = 'tsui.tony@gmail.com'
 # List of additional contributors. Format: 'Name Email'
-CONTRIBUTORS = []
+CONTRIBUTORS = [ 'Tango Desktop Project http://tango-project.org' ]
 WEBSITE = 'http://guppy.nongnu.org'
-COPYRIGHT = 'Copyright (C) 2005 Tony Tsui'
-LICENSE = 'GNU Public License'
+COPYRIGHT = 'Copyright (C) 2005-2006 Tony Tsui'
+LICENSE = 'GNU Public License & CC-by-SA 2.5 License'
