@@ -174,6 +174,19 @@ class Puppy:
 			status = self.popen_obj.poll()
 			
 		return status
+
+	def isActive(self):
+		lock_file = open(Puppy.LOCK_FILE, 'a')
+		
+		try:
+			fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
+			result = False
+		except:
+			result = True
+			
+		lock_file.close()
+	
+		return result
 		
 	def listDir(self, path=None):
 		args = ['-c', 'dir']
@@ -193,12 +206,22 @@ class Puppy:
 		
 		listing = []
 		# Parse output of output_file and return it as a list
+		space = ' '
 		for line in output:
 			entry = line.split()
-			space = ' '
-			item = [ entry[0], space.join(entry[7:]),
-			         "%s %s %s %s" % (entry[2], entry[3], entry[4], entry[6]),
-			         entry[1] ]
+
+			# Convert from Toppy's code page of ISO-8859-1 to UTF-8
+			# FIXME: Toppy may not always be ISO-8859-1. How do we find out
+			#        what encoding to use?
+			filename = unicode(space.join(entry[7:]), 'iso8859-1', 'replace')
+			filename = filename.encode('utf-8')
+
+			date = unicode("%s %s %s %s" % (entry[2], entry[3], entry[4],
+			                                entry[6]),
+			               'iso8859-1', 'replace')
+			date = date.encode('utf-8')
+
+			item = [ entry[0], filename, date, entry[1] ]
 			listing.append(item)
 		
 		return listing
@@ -285,26 +308,13 @@ class Puppy:
 		
 		return
 			
-	def _anotherPuppyActive(self):
-		lock_file = open(Puppy.LOCK_FILE, 'a')
-		
-		try:
-			fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
-			result = False
-		except:
-			result = True
-			
-		lock_file.close()
-	
-		return result
-		
 	def _execute(self, args_list):
 		cmd = [ self.cmd ] + args_list
 		
 		if DEBUG:
 			print 'cmd = ', cmd
 
-		if self._anotherPuppyActive():
+		if self.isActive():
 			raise PuppyBusyError('Can not get exclusive lock on ' + Puppy.LOCK_FILE)
 			
 		self.popen_obj = popen2.Popen4(cmd)
